// dharma.c by Kosei Demura 2006-6-4
// My web site is http://demura.net
// This program uses the Open Dynamics Engine (ODE) by Russell Smith.
// The ODE web site is http://ode.org/
#include <ode/ode.h>
#include <drawstuff/drawstuff.h>

#ifdef dDOUBLE
#define dsDrawSphere dsDrawSphereD
#endif

#define MAX_CONTACTS 4

static dWorldID world;
static dSpaceID space; 
static dGeomID  ground;
static dJointGroupID contactgroup; 
dsFunctions fn;

typedef struct {    
  dBodyID body;     
  dGeomID geom;     
  dReal   radius;
  dReal   length;
  dReal   mass;
} MyLink;

MyLink trans_ball,ball;  
dGeomID  encap_ball_geom;   // encapsulated geometries
dJointID joint;

static void nearCallback (void *data, dGeomID o1, dGeomID o2)
{
  int i;

  // exit without doing anything if the two bodies are connected by a joint             
  dBodyID b1 = dGeomGetBody(o1);
  dBodyID b2 = dGeomGetBody(o2);
  if (b1 && b2 && dAreConnectedExcluding (b1,b2,dJointTypeContact)) return;

  dContact contact[MAX_CONTACTS];   // up to MAX_CONTACTS contacts per box-box          
  for (i=0; i<MAX_CONTACTS; i++) {
    contact[i].surface.mode = dContactBounce | dContactSoftCFM;
    contact[i].surface.mu = 5; //dInfinity;
    //contact[i].surface.mu2 = dInfinity;
    contact[i].surface.bounce = 0.01;
    contact[i].surface.bounce_vel = 0.01;
    contact[i].surface.soft_cfm = 0.00001;
  }
  if (int numc = dCollide (o1,o2,MAX_CONTACTS,&contact[0].geom,
                           sizeof(dContact))) {
    for (i=0; i<numc; i++) {
      dJointID c = dJointCreateContact (world,contactgroup,contact+i);
      dJointAttach (c,b1,b2);
    }
  }
}

static void drawGeom(dGeomID g, const dReal *pos, const dReal *R)
{
  if (!g) return;
  if (!pos) pos = dGeomGetPosition(g);
  if (!R) R = dGeomGetRotation(g);

  int type = dGeomGetClass(g);

  if (type == dCCylinderClass) {
    dReal radius,length;
    dGeomCCylinderGetParams(g,&radius,&length);
    dsDrawCappedCylinderD(pos,R,length,radius);
  }
  else if (type == dSphereClass) {
    dsDrawSphereD(pos,R,dGeomSphereGetRadius (g));
  }
  else if (type == dGeomTransformClass) {
    dGeomID g2 = dGeomTransformGetGeom(g);
    const dReal *pos2 = dGeomGetPosition(g2);
    const dReal *R2   = dGeomGetRotation(g2);
    dVector3 actual_pos;
    dMatrix3 actual_R;
    dMULTIPLY0_331(actual_pos,R,pos2);
    actual_pos[0] += pos[0];
    actual_pos[1] += pos[1];
    actual_pos[2] += pos[2];
    dMULTIPLY0_333 (actual_R,R,R2);
    drawGeom(g2,actual_pos,actual_R);
  }
}

static void simLoop (int pause)
{
  dSpaceCollide(space,0,&nearCallback); 
  dWorldStep(world,0.01);
  dJointGroupEmpty(contactgroup);
  dsSetColor(1.0,0.0,0.0);
  drawGeom(trans_ball.geom, 0, 0);
  drawGeom(ball.geom, 0, 0);
}

void start()
{
  static float xyz[3] = {   3.5,0.0,1.0};
  static float hpr[3] = {-180.0,0.0,0.0};
  dsSetSphereQuality(3);
  dsSetViewpoint (xyz,hpr);
}

// Create an object 
void createDaruma() {
  dMass m,m1;
  dReal x0 = 0.0, y0 = 0.0, z0 = 1.0;
  dReal dx = 0.0, dy = 0.0, dz = 0.4;  // delta position

  trans_ball.geom = dCreateGeomTransform(space); 
  trans_ball.body = dBodyCreate(world);
  dGeomTransformSetCleanup(trans_ball.geom,1);

  // Body: COG transformed ball
  trans_ball.radius = 0.4;
  trans_ball.mass   = 10.0;
  dMassSetZero(&m);
  dMassSetSphereTotal(&m,trans_ball.mass,trans_ball.radius);
  encap_ball_geom = dCreateSphere(0,trans_ball.radius); 
  // attach encap_ball to trans_ball
  dGeomTransformSetGeom(trans_ball.geom,encap_ball_geom); 

  dGeomSetPosition(encap_ball_geom, dx, dy, dz);
  dMassTranslate(&m, dx, dy, dz);

  dGeomSetBody(trans_ball.geom,trans_ball.body);       
  dBodySetMass(trans_ball.body,&m);            
  dGeomSetPosition(trans_ball.geom, x0 -dx, y0 -dy, z0 - dz);

  // Head
  ball.radius = 0.3;
  ball.mass   = 0.001;
  ball.body   = dBodyCreate(world);   
  dMassSetZero(&m1);
  dMassSetSphereTotal(&m1,ball.mass, ball.radius);
  dBodySetMass(ball.body,&m1);            
  dBodySetPosition(ball.body, x0, y0, z0 + trans_ball.radius + ball.radius);
  ball.geom = dCreateSphere(space,ball.radius);
  dGeomSetBody(ball.geom,ball.body);      

  // fixed joint
  joint = dJointCreateFixed(world, 0);
  dJointAttach(joint, trans_ball.body,ball.body);
  dJointSetFixed(joint); 
}


void command(int cmd)
{
  switch(cmd) {
  case 'f':
    dBodyAddTorque(ball.body,  100, 0, 0); break;
  case 'j':
    dBodyAddTorque(ball.body, -100, 0, 0); break;
  }
}

void  setDrawStuff() {
  fn.version = DS_VERSION;
  fn.start   = &start;
  fn.step    = &simLoop;
  fn.command = &command;
  fn.stop    = NULL;
  fn.path_to_textures = "../../drawstuff/textures";
}

int main (int argc, char **argv)
{
  setDrawStuff();

  world = dWorldCreate();
  space = dHashSpaceCreate(0);
  contactgroup = dJointGroupCreate(0); 

  dWorldSetGravity(world,0,0,-9.8);
  ground = dCreatePlane(space,0,0,1,0);
  createDaruma();

  dsSimulationLoop (argc,argv,640,480,&fn);

  dWorldDestroy (world);
  
  return 0;
}
